// DemodDecimationAVX.cpp : This file contains the 'main' function. Program execution begins and ends there.
//
#pragma warning(disable : 4514 4701 4710 4711 4820 5045)

#define _CRT_SECURE_NO_WARNINGS
#define LIGO_BIQUAD
#define CHECK_RESULT
#define MAX_WRITE_RESULT 1000
#include <chrono>
#include <array>
#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include <string.h>
#include "coeff_traits.h"
#include "instructionset.h"
#include "mxcsr.h"
#include "Demodulation.h"

const double twopi = 6.2831853071795864769;

const int dataN = 1000;
const int stride = 32;
const int samples = 8;
const int alignment = 64;
const int sections = 3;

#ifdef _DEBUG
const int repeattest = 1;
#else
const int repeattest = 10000;
#endif

// Coefficient array for 3 second order section (6th order IIR filter)
using coeff_traits3 = coeff_traits<3>;
using coeff_order_6 = coeff_traits3::coeff_array;
using coeff_order_6_wide_4 = coeff_traits3::coeff_wide_array<4>;
using coeff_order_6_wide_8 = coeff_traits3::coeff_wide_array<8>;


//static const double coeffdec[13] = { 0.00162185538923,
//							  0.73342532779703, -0.02862365091314, 1.44110125961504,  1.67905228090487,
//							  0.77657563380963, -0.08304311675394, 0.18851328163424,  0.32889453107067,
//							  0.83213081484618, -0.12573495191273, 0.0940911979108501,0.13622543115194 };

// cheby2("LowPass",6,60,8000)
// rate 524288 Hz
//static const coeff_order_6 chebydirect = { 9.410818796785352805325808e-04,
//							 -1.9752147772326503,  0.9777400275124176, -1.9901574406868523,  1.0000000000000000,
//							 -1.9296605059414182,  0.9324647855643438, -1.9816725895662000,  1.0000000000000000,
//							 -1.8920342775852999,  0.8952193075414041, -1.8671374119651403,  0.9999999999999999 };

// cheby2("LowPass",6,60,32000)
// rate 524288 Hz
static coeff_order_6 chebydirect = { 1.159664984828962224633342e-03,
							 -1.8740625376565518,  0.9136862817567704, -1.8446997352645189,  1.0000000000000000,
							 -1.7133654266738769,  0.7545999972582751, -1.7196326892723282,  1.0000000000000000,
							 -1.5936647899894101,  0.6381719943234752, -0.5598252720330773,  1.0000000000000000 };

static coeff_order_6_wide_4 chebydirect4{ coeff_traits3::wide(chebydirect, chebydirect, chebydirect, chebydirect) };
static coeff_order_6_wide_8 chebydirect8{ coeff_traits3::wide(chebydirect, chebydirect, chebydirect, chebydirect, chebydirect, chebydirect, chebydirect, chebydirect) };

static coeff_order_6 chebybiquad = coeff_traits3::biquad_shuffle(chebydirect);
static coeff_order_6_wide_4 chebybiquad4{ coeff_traits3::wide(chebybiquad, chebybiquad, chebybiquad, chebybiquad) };
static coeff_order_6_wide_8 chebybiquad8{ coeff_traits3::wide(chebybiquad, chebybiquad, chebybiquad, chebybiquad, chebybiquad, chebybiquad, chebybiquad, chebybiquad) };

#ifdef LIGO_BIQUAD
const double* coeff = chebybiquad.data();
const double* coeff4 = chebybiquad4.data();
const double* coeff8 = chebybiquad8.data();
#else
const double* coeff = chebydirect.data();
const double* coeff4 = chebydirect4.data();
const double* coeff8 = chebydirect8.data();
#endif

#ifdef __GNUC__
static const char* const dirpath = R"()";
#else
static const char* const dirpath = R"(C:\SlowControls\EPICS\Utilities\BiquadAVX\)";
#endif

static
void write_inputs(double* out, int n, const char* filename, int offset = 1)
{
	int max = n;
	if (max > dataN) max = dataN;
	if (max <= 0) return;
	if (offset < 0) return;
	if (!filename) return;

	char filen[1024];
#ifdef __GNUC__
	if ((strlen(filename) >= 1) && ((filename[0] == '~') || (filename[0] == '/'))) {
#else
	if ((strlen(filename) >= 2) && (filename[1] == ':')) {
#endif
		strncpy(filen, filename, sizeof(filen));
		filen[sizeof(filen) - 1] = 0;
	}
	else {
		strncpy(filen, dirpath, sizeof(filen));
		filen[sizeof(filen) - 1] = 0;
		size_t len = strlen(filen);
		if (len < sizeof(filen)) strncpy(filen + len, filename, sizeof(filen) - len);
		filen[sizeof(filen) - 1] = 0;
	}

	FILE* fout = fopen(filen, "w");
	if (!fout) return;

	for (int i = 0; i < max; ++i)
	{
		fprintf(fout, "%4i %12.9f %12.9f %12.9f %12.9f %12.9f %12.9f %12.9f %12.9f\n", i * offset,
			out[i * stride + 0], out[i * stride + 4], out[i * stride + 8], out[i * stride + 12],
			out[i * stride + 16], out[i * stride + 20], out[i * stride + 24], out[i * stride + 28]);
	}
	fclose(fout);
}

static
void write_results(double* out, int n, const char* filename, int offset = samples)
{
	int max = n;
	if (max > dataN) max = dataN;
	if (max <= 0) return;
	if (offset < 0) return;
	if (!filename) return;

	char filen[1024];
#ifdef __GNUC__
	if ((strlen(filename) >= 1) && ((filename[0] == '~') || (filename[0] == '/'))) {
#else
	if ((strlen(filename) >= 2) && (filename[1] == ':')) {
#endif
		strncpy(filen, filename, sizeof(filen));
		filen[sizeof(filen) - 1] = 0;
	}
	else {
		strncpy(filen, dirpath, sizeof(filen));
		filen[sizeof(filen) - 1] = 0;
		size_t len = strlen(filen);
		if (len < sizeof(filen)) strncpy(filen + len, filename, sizeof(filen) - len);
		filen[sizeof(filen) - 1] = 0;
	}

	FILE* fout = fopen(filen, "w");
	if (!fout) return;

	for (int i = 0; i < max; ++i)
	{
		fprintf(fout, "%4i %12.9f %12.9f %12.9f %12.9f %12.9f %12.9f %12.9f %12.9f %12.9f %12.9f %12.9f %12.9f %12.9f %12.9f %12.9f %12.9f\n", i * offset,
			out[i * 2 * stride +  0], out[i * 2 * stride +  1], out[i * 2 * stride +  8], out[i * 2 * stride +  9],
			out[i * 2 * stride + 16], out[i * 2 * stride + 17], out[i * 2 * stride + 24], out[i * 2 * stride + 25],
			out[i * 2 * stride + 32], out[i * 2 * stride + 33], out[i * 2 * stride + 40], out[i * 2 * stride + 41],
			out[i * 2 * stride + 48], out[i * 2 * stride + 49], out[i * 2 * stride + 56], out[i * 2 * stride + 57]);
	}
	fclose(fout);
}

#ifdef CHECK_RESULT
static
double check_result(const double* inp1, const double* inp2, int n)
{
	double max = 0;
	for (int i = 0; i < n; ++i)
	{
		double diff = abs(inp1[i] - inp2[i]);
		if (isnan(inp1[i]) || isnan(inp2[i])) diff = 1E9;
		if (diff > max) max = diff;
	}
	return max;
}
#endif

//#include <format>
//#include <random>

int main()
{
	//int m = sizeof(bool); 
	//std::atomic<bool> bb{};
	//bool dd = bb.is_always_lock_free;
	//int mm = sizeof(bb);
	//auto s1 = std::format("check {}", m);
	//const char* fmt = "JJJ {}";
	//auto s2 = std::format(fmt, m);
	//
	//using gpsclock = std::chrono::gps_clock;
	//auto gpsnow = gpsclock::now();
	//std::knuth_b generator (gpsnow.time_since_epoch().count());
	//for (int i = 0; i < 5; ++i) {
	//	using gpsclock = std::chrono::gps_clock;
	//	auto gpsnow = gpsclock::now();
	//	int64_t gnow = 100 * gpsnow.time_since_epoch().count() + (generator()- generator.min()) / (generator.max()/100);
	//	printf("GPS now in nsec  %lli\n", gnow);
	//	printf("GPS now in usec %18.9f\n", (double)(gnow) / 1E9);
	//	int64_t us = (gnow + 499) / 1000;
	//	printf("GPS now in usec  %lli\n", us);
	//	double x = (double)(us) / 1E6;
	//	printf("GPS now in usec %15.6f\n", x);
	//	int64_t rnow = llround(1E6 * x);
	//	printf("GPS now in usec  %lli\n\n", rnow);
	//}

	using clock = std::chrono::system_clock;
	using sec = std::chrono::duration<double>;

	printf("DEMODULATION/FILTER EXECUTATION TIME COMPARISON\n");
	InstructionSet::PRINTAVX();
	fz_daz();
	print_mxcsr();

	double* inp = (double*)_mm_malloc(dataN * stride * samples * sizeof(double), alignment);
	double* out = (double*)_mm_malloc(2 * dataN * stride * sizeof(double), alignment);
	double* hist = (double*)_mm_malloc(4 * stride * sections * sizeof(double), alignment);
	if (!inp || !out || !hist) return 1;

	// ---------------------------------------------------------------------------------------------
	// initialize input
	for (int i = 0; i < dataN * samples; ++i)
	{
		for (int j = 0; j < stride; ++j)
		{
			if (inp) inp[i * stride + j] = sin(twopi * i * 0.01 * (j + 1));
		}
	}
	write_inputs(inp, MAX_WRITE_RESULT, R"(file_demodulation_inp.txt)");

	// ---------------------------------------------------------------------------------------------
	// zero history
	printf("\ndemodulation_decimation_stride8_section3_std:\n");
	if (out) memset(out, 0, 2 * dataN * stride * sizeof(double));
	if (hist) memset(hist, 0, 4 * stride * sections * sizeof(double));
	avx_stride_struct* avxstridehist = (avx_stride_struct*)_mm_malloc(sizeof(avx_stride_struct), alignment);
	avx_stride_struct_init(avxstridehist, 1.0);
	double freq = 0.01 * 25.1;
	avx_stride_struct_set_frequency(avxstridehist, freq);

	// time demodulation_decimation
	auto before = clock::now();
	for (int k = 0; k < repeattest; ++k)
	{
		for (int i = 0; i < dataN; ++i)
		{
			for (int d = 0; d < samples; ++d)
			{
				demodulation_decimation_stride8_section3_std(inp + (i * samples + d) * stride, freq, out + 2 * (i * stride), coeff, avxstridehist, stride);
			}
		}
	}
	sec duration_demodulation_decimation = clock::now() - before;
	printf("It took %g sec\n", duration_demodulation_decimation.count());

	// write results
	write_results(out, MAX_WRITE_RESULT, R"(file_demodulation_decimation_stride8_section3_std.txt)");
#ifdef CHECK_RESULT
	double* outfirst = out;
	out = (double*)_mm_malloc(2 * dataN * stride * sizeof(double), alignment);
#endif

#if USE_SSE3 || defined(_MSC_VER)
	// ---------------------------------------------------------------------------------------------
	// zero history
	printf("\ndemodulation_decimation_stride8_section3_sse3:\n");
	if (out) memset(out, 0, 2 * dataN * stride * sizeof(double));
	if (hist) memset(hist, 0, 4 * stride * sections * sizeof(double));
	avx_stride_struct_init(avxstridehist, 1.0);
	avx_stride_struct_set_frequency(avxstridehist, freq);

	// init cordic
	// time demodulation_decimation_stride8_section3_sse3
	before = clock::now();
	for (int k = 0; k < repeattest; ++k)
	{
		for (int i = 0; i < dataN; ++i)
		{
			for (int d = 0; d < samples; ++d)
			{
				demodulation_decimation_stride8_section3_sse3(inp + (i * samples + d) * stride, freq, out + 2 * i * stride, coeff, avxstridehist, stride);
			}
		}
	}
	sec duration_demodulation_decimation_asse3_section3 = clock::now() - before;
	printf("It took %g sec\n", duration_demodulation_decimation_asse3_section3.count());
	printf("It was %gx faster\n", duration_demodulation_decimation.count() / duration_demodulation_decimation_asse3_section3.count());
#ifdef CHECK_RESULT
	double diff = check_result(out, outfirst, 2 * dataN * stride);
	printf("Biggest deviation %g\n", diff);
#endif

	// write results
	write_results(out, MAX_WRITE_RESULT, R"(file_demodulation_decimation_stride8_section3_sse3.txt)");
#endif

#if defined(__AVX2__) || defined(_MSC_VER)
	// ---------------------------------------------------------------------------------------------
	// zero history
	printf("\ndemodulation_decimation_stride8_section3_avx2:\n");
	if (out) memset(out, 0, 2 * dataN * stride * sizeof(double));
	if (hist) memset(hist, 0, 4 * stride * sections * sizeof(double));
	avx_stride_struct_init(avxstridehist, 1.0);
	avx_stride_struct_set_frequency(avxstridehist, freq);

	// init cordic
	// time demodulation_decimation_avx2_stride8_section3
	before = clock::now();
	for (int k = 0; k < repeattest; ++k)
	{
		for (int i = 0; i < dataN; ++i)
		{
			for (int d = 0; d < samples; ++d)
			{
				demodulation_decimation_stride8_section3_avx2(inp + (i * samples + d) * stride, freq, out + 2 * i * stride, coeff, avxstridehist, stride);
			}
		}
	}
	sec duration_demodulation_decimation_avx2_section3 = clock::now() - before;
	printf("It took %g sec\n", duration_demodulation_decimation_avx2_section3.count());
	printf("It was %gx faster\n", duration_demodulation_decimation.count() / duration_demodulation_decimation_avx2_section3.count());
#ifdef CHECK_RESULT
	diff = check_result(out, outfirst, 2 * dataN * stride);
	printf("Biggest deviation %g\n", diff);
#endif

	// write results
	write_results(out, MAX_WRITE_RESULT, R"(file_demodulation_decimation_stride8_section3_avx2.txt)");
#endif

#if (defined(__AVX512F__) && defined(__AVX512CD__) && defined(__AVX512DQ__) && defined(__AVX512BW__)) || defined(_MSC_VER)
	if (InstructionSet::AVX512F())
	{
		// ---------------------------------------------------------------------------------------------
		// zero history
		printf("\ndemodulation_decimation_stride8_section3_avx512:\n");
		if (out) memset(out, 0, 2 * dataN * stride * sizeof(double));
		if (hist) memset(hist, 0, 4 * stride * sections * sizeof(double));
		//avx_stride_struct* avxstridehist = (avx_stride_struct*)_mm_malloc(sizeof(avx_stride_struct), alignment);
		avx_stride_struct_init(avxstridehist, 1.0);
		avx_stride_struct_set_frequency(avxstridehist, freq);

		// init cordic
		// time demodulation_decimation_avx2_stride8_section3
		before = clock::now();
		for (int k = 0; k < repeattest; ++k)
		{
			for (int i = 0; i < dataN; ++i)
			{
				for (int d = 0; d < samples; ++d)
				{
					demodulation_decimation_stride8_section3_avx512(inp + (i * samples + d) * stride, freq, out + 2 * i * stride, coeff, avxstridehist, stride);
				}
			}
		}
		sec duration_demodulation_decimation_avx512_section3 = clock::now() - before;
		printf("It took %g sec\n", duration_demodulation_decimation_avx512_section3.count());
		printf("It was %gx faster\n", duration_demodulation_decimation.count() / duration_demodulation_decimation_avx512_section3.count());
#ifdef CHECK_RESULT
		diff = check_result(out, outfirst, 2 * dataN * stride);
		printf("Biggest deviation %g\n", diff);
#endif

		// write results
		write_results(out, MAX_WRITE_RESULT, R"(file_demodulation_decimation_stride8_section3_avx512.txt)");
	}
#endif

	// ---------------------------------------------------------------------------------------------
	// zero history
	printf("\ndemodulation_decimation_rotation8_section3_std:\n");
	if (out) memset(out, 0, 2 * dataN * stride * sizeof(double));
	if (hist) memset(hist, 0, 4 * stride * sections * sizeof(double));
	avx_rotation_struct* avxrothist = (avx_rotation_struct*)_mm_malloc(sizeof(avx_rotation_struct), alignment);
	avx_rotation_struct_init(avxrothist, 1.0);
	double freqarr[stride] = { };
	for (unsigned int j = 0; j < stride; ++j) 
	{
		freqarr[j] = 0.01 * (25.0 + (j - 16.0) * 0.1);
		avx_rotation_struct_set_frequency(avxrothist, freqarr[j], j);
	}

	// time demodulation_decimation_rotation8_section3_std
	before = clock::now();
	for (int k = 0; k < repeattest; ++k)
	{
		for (int i = 0; i < dataN; ++i)
		{
			for (int d = 0; d < samples; ++d)
			{
				demodulation_decimation_rotation8_section3_std(inp[(i * samples + d) * stride], freqarr, out + 2 * (i * stride), coeff, avxrothist, stride);
			}
		}
	}
	sec duration_demodulation_decimation_rot8 = clock::now() - before;
	printf("It took %g sec\n", duration_demodulation_decimation_rot8.count());

	// write results
	write_results(out, MAX_WRITE_RESULT, R"(file_demodulation_decimation_rotation8_section3_std.txt)");
#ifdef CHECK_RESULT
	_mm_free(outfirst);
	outfirst = out;
	out = (double*)_mm_malloc(2 * dataN * stride * sizeof(double), alignment);
#endif

#if USE_SSE3
	// ---------------------------------------------------------------------------------------------
	// zero history
	printf("\ndemodulation_decimation_rotation8_section3_sse3:\n");
	if (out) memset(out, 0, 2 * dataN * stride * sizeof(double));
	if (hist) memset(hist, 0, 4 * stride * sections * sizeof(double));
	// init cordic
	avx_rotation_struct_init(avxrothist, 1.0);
	for (unsigned int j = 0; j < stride; ++j)
	{
		freqarr[j] = 0.01 * (25.0 + (j - 16.0) * 0.1);
		avx_rotation_struct_set_frequency(avxrothist, freqarr[j], j);
	}

	// time demodulation_decimation_rotation8_section3_sse3
	before = clock::now();
	for (int k = 0; k < repeattest; ++k)
	{
		for (int i = 0; i < dataN; ++i)
		{
			for (int d = 0; d < samples; ++d)
			{
				demodulation_decimation_rotation8_section3_sse3(inp[(i * samples + d) * stride], freqarr, out + 2 * (i * stride), coeff, avxrothist, stride);
			}
		}
	}
	sec duration_demodulation_decimation_rot8_sse3_section3 = clock::now() - before;
	printf("It took %g sec\n", duration_demodulation_decimation_rot8_sse3_section3.count());
	printf("It was %gx faster\n", duration_demodulation_decimation_rot8.count() / duration_demodulation_decimation_rot8_sse3_section3.count());
#ifdef CHECK_RESULT
	diff = check_result(out, outfirst, 2 * dataN * stride);
	printf("Biggest deviation %g\n", diff);
#endif

	// write results
	write_results(out, MAX_WRITE_RESULT, R"(file_demodulation_decimation_rotation8_section3_sse3.txt)");
#endif

#if defined(__AVX2__) || defined(_MSC_VER)
	// ---------------------------------------------------------------------------------------------
	// zero history
	printf("\ndemodulation_decimation_rotation8_section3_avx2:\n");
	if (out) memset(out, 0, 2 * dataN * stride * sizeof(double));
	if (hist) memset(hist, 0, 4 * stride * sections * sizeof(double));
	// init cordic
	avx_rotation_struct_init(avxrothist, 1.0);
	for (unsigned int j = 0; j < stride; ++j)
	{
		freqarr[j] = 0.01 * (25.0 + (j - 16.0) * 0.1);
		avx_rotation_struct_set_frequency(avxrothist, freqarr[j], j);
	}

	// time demodulation_decimation_rotation8_section3_avx2
	before = clock::now();
	for (int k = 0; k < repeattest; ++k)
	{
		for (int i = 0; i < dataN; ++i)
		{
			for (int d = 0; d < samples; ++d)
			{
				demodulation_decimation_rotation8_section3_avx2(inp[(i * samples + d) * stride], freqarr, out + 2 * (i * stride), coeff, avxrothist, stride);
			}
		}
	}
	sec duration_demodulation_decimation_rot8_avx2_section3 = clock::now() - before;
	printf("It took %g sec\n", duration_demodulation_decimation_rot8_avx2_section3.count());
	printf("It was %gx faster\n", duration_demodulation_decimation_rot8.count() / duration_demodulation_decimation_rot8_avx2_section3.count());
#ifdef CHECK_RESULT
	diff = check_result(out, outfirst, 2 * dataN * stride);
	printf("Biggest deviation %g\n", diff);
#endif

	// write results
	write_results(out, MAX_WRITE_RESULT, R"(file_demodulation_decimation_rotation8_section3_avx2.txt)");
#endif

#if (defined(__AVX512F__) && defined(__AVX512CD__) && defined(__AVX512DQ__) && defined(__AVX512BW__)) || defined(_MSC_VER)
	if (InstructionSet::AVX512F())
	{
		// ---------------------------------------------------------------------------------------------
		// zero history
		printf("\ndemodulation_decimation_rotation8_section3_avx512:\n");
		if (out) memset(out, 0, 2 * dataN * stride * sizeof(double));
		if (hist) memset(hist, 0, 4 * stride * sections * sizeof(double));
		// init cordic
		avx_rotation_struct_init(avxrothist, 1.0);
		for (unsigned int j = 0; j < stride; ++j)
		{
			freqarr[j] = 0.01 * (25.0 + (j - 16.0) * 0.1);
			avx_rotation_struct_set_frequency(avxrothist, freqarr[j], j);
		}

		// time demodulation_decimation_rotation8_section3_avx512
		before = clock::now();
		for (int k = 0; k < repeattest; ++k)
		{
			for (int i = 0; i < dataN; ++i)
			{
				for (int d = 0; d < samples; ++d)
				{
					demodulation_decimation_rotation8_section3_avx512(inp[(i * samples + d) * stride], freqarr, out + 2 * (i * stride), coeff, avxrothist, stride);
				}
			}
		}
		sec duration_demodulation_decimation_rot8_avx512_section3 = clock::now() - before;
		printf("It took %g sec\n", duration_demodulation_decimation_rot8_avx512_section3.count());
		printf("It was %gx faster\n", duration_demodulation_decimation_rot8.count() / duration_demodulation_decimation_rot8_avx512_section3.count());
#ifdef CHECK_RESULT
		diff = check_result(out, outfirst, 2 * dataN * stride);
		printf("Biggest deviation %g\n", diff);
#endif

		// write results
		write_results(out, MAX_WRITE_RESULT, R"(file_demodulation_decimation_rotation8_section3_avx512.txt)");
	}
#endif
}

