#define _CRT_SECURE_NO_WARNINGS
#define _USE_MATH_DEFINES
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <assert.h>
#include <stdalign.h>
#include "mxcsr.h"
#include "Biquad.h"
#include "Demodulation.h"

// ---------------------------------------------------------------------------------------------
// Constants for test execution
#define dataN 1000
#define stride 32
#define samples 8
#define alignment 64
#define sections 3

#ifdef _DEBUG
#define repeattest 1
#else
#define repeattest 1000
#endif

#define MAX_WRITE_RESULT 0

#define demodfreq (0.01 * 25.1)
#define tolerance (sqrt(sqrt((double)repeattest)) * 2E-15)

// ---------------------------------------------------------------------------------------------
// directory path for results
#ifdef __GNUC__
#define dirpath "~/"
#else
#define dirpath "C:\\SlowControls\\EPICS\\Utilities\\BiquadAVX\\)"
#endif

// ---------------------------------------------------------------------------------------------
// filter coefficients
static const double coeffdec[13] = { 0.00162185538923,
							  0.73342532779703, -0.02862365091314, 1.44110125961504,  1.67905228090487,
							  0.77657563380963, -0.08304311675394, 0.18851328163424,  0.32889453107067,
							  0.83213081484618, -0.12573495191273, 0.0940911979108501,0.13622543115194 };

alignas (16) static double coeffdec2[2 * 13];
alignas (32) static double coeffdec4[4 * 13];
alignas (64) static double coeffdec8[8 * 13];

// ---------------------------------------------------------------------------------------------
// demodulation frequencies
double demodfreqarr[stride] = { 0 };

// ---------------------------------------------------------------------------------------------
// check_result
static
double check_result(const double* inp1, const double* inp2, int n)
{
	double max = 0;
	for (int i = 0; i < n; ++i)
	{
		double diff = fabs(inp1[i] - inp2[i]);
		if (isnan(inp1[i]) || isnan(inp2[i])) diff = 1E9;
		if (diff > max) max = diff;
	}
	return max;
}

// ---------------------------------------------------------------------------------------------
// write_results
static
void write_results(double* out, int n, const char* filename, int offset)
{
	int max = n;
	if (max > dataN) max = dataN;
	if (max <= 0) return;
	if (offset < 0) return;
	if (!filename) return;

	char filen[1024];
#ifdef __GNUC__
	if ((strlen(filename) >= 1) && ((filename[0] == '~') || (filename[0] == '/'))) {
#else
	if ((strlen(filename) >= 2) && (filename[1] == ':')) {
#endif
		strncpy(filen, filename, sizeof(filen));
		filen[sizeof(filen) - 1] = 0;
	}
	else {
		strncpy(filen, dirpath, sizeof(filen));
		filen[sizeof(filen) - 1] = 0;
		size_t len = strlen(filen);
		if (len < sizeof(filen)) strncpy(filen + len, filename, sizeof(filen) - len);
		filen[sizeof(filen) - 1] = 0;
	}

	FILE* fout = fopen(filename, "w");
	if (!fout) return;

	for (int i = 0; i < max; ++i)
	{
		fprintf(fout, "%4i %12.9f %12.9f %12.9f %12.9f %12.9f %12.9f %12.9f %12.9f\n", i * offset,
			out[i * stride + 0], out[i * stride + 4], out[i * stride + 8], out[i * stride + 12],
			out[i * stride + 16], out[i * stride + 20], out[i * stride + 24], out[i * stride + 28]);
	}
	fclose(fout);
}

// ---------------------------------------------------------------------------------------------
// write_demod_results
static
void write_demod_results(double* out, int n, const char* filename, int offset)
{
	int max = n;
	if (max > dataN) max = dataN;
	if (max <= 0) return;
	if (offset < 0) return;
	if (!filename) return;

	char filen[1024];
#ifdef __GNUC__
	if ((strlen(filename) >= 1) && ((filename[0] == '~') || (filename[0] == '/'))) {
#else
	if ((strlen(filename) >= 2) && (filename[1] == ':')) {
#endif
		strncpy(filen, filename, sizeof(filen));
		filen[sizeof(filen) - 1] = 0;
	}
	else {
		strncpy(filen, dirpath, sizeof(filen));
		filen[sizeof(filen) - 1] = 0;
		size_t len = strlen(filen);
		if (len < sizeof(filen)) strncpy(filen + len, filename, sizeof(filen) - len);
		filen[sizeof(filen) - 1] = 0;
	}

	FILE* fout = fopen(filen, "w");
	if (!fout) return;

	for (int i = 0; i < max; ++i)
	{
		fprintf(fout, "%4i %12.9f %12.9f %12.9f %12.9f %12.9f %12.9f %12.9f %12.9f %12.9f %12.9f %12.9f %12.9f %12.9f %12.9f %12.9f %12.9f\n", i * offset,
			out[i * 2 * stride + 0], out[i * 2 * stride + 1], out[i * 2 * stride + 8], out[i * 2 * stride + 9],
			out[i * 2 * stride + 16], out[i * 2 * stride + 17], out[i * 2 * stride + 24], out[i * 2 * stride + 25],
			out[i * 2 * stride + 32], out[i * 2 * stride + 33], out[i * 2 * stride + 40], out[i * 2 * stride + 41],
			out[i * 2 * stride + 48], out[i * 2 * stride + 49], out[i * 2 * stride + 56], out[i * 2 * stride + 57]);
	}
	fclose(fout);
}

// ---------------------------------------------------------------------------------------------
// test_setup
_Bool test_setup(double** inp, double** out, double** outfirst, double** hist)
{
	fz_daz();
	print_mxcsr();
	if ((inp == NULL) || (out == NULL) || (outfirst == NULL) || (hist == NULL)) {
		return 0;
	}

	// ---------------------------------------------------------------------------------------------
	// fill coeffdecX
	for (int i = 0; i < 13; ++i)
	{
		coeffdec2[2 * i + 1] = coeffdec2[2 * i + 0] = coeffdec[i];
		coeffdec4[4 * i + 3] = coeffdec4[4 * i + 2] = coeffdec4[4 * i + 1] = coeffdec4[4 * i + 0] = coeffdec[i];
		coeffdec8[8 * i + 7] = coeffdec8[8 * i + 6] = coeffdec8[8 * i + 5] = coeffdec8[8 * i + 4] = coeffdec[i];
		coeffdec8[8 * i + 3] = coeffdec8[8 * i + 2] = coeffdec8[8 * i + 1] = coeffdec8[8 * i + 0] = coeffdec[i];
	}

	// ---------------------------------------------------------------------------------------------
	// allocate memory
	*inp = (double*)_mm_malloc(dataN * stride * samples * sizeof(double), alignment);
	*out = (double*)_mm_malloc(2 * dataN * stride * sizeof(double), alignment);
	*outfirst = (double*)_mm_malloc(2 * dataN * stride * sizeof(double), alignment);
	*hist = (double*)_mm_malloc(4 * stride * sections * sizeof(double), alignment);
	if ((*inp == NULL) || (*out == NULL) || (*outfirst == NULL) || (*hist == NULL)) {
		return 0;
	}

	// ---------------------------------------------------------------------------------------------
	// initialize input
	for (int i = 0; i < dataN * samples; ++i)
	{
		for (int j = 0; j < stride; ++j)
		{
			(*inp)[i * stride + j] = sin(2.0 * M_PI * i * 0.01 * (j + 1));
		}
	}
	write_results(*inp, MAX_WRITE_RESULT, "test_inp.txt", 1);

	// ---------------------------------------------------------------------------------------------
	// zero history
	printf("\ngenerate reference data (iir_filter_biquad)\n");
	memset(*outfirst, 0, dataN * stride * sizeof(double));
	memset(*hist, 0, 2 * stride * sections * sizeof(double));

	// ---------------------------------------------------------------------------------------------
	// fill reference data by using standard biquad
	for (int k = 0; k < repeattest; ++k) {
		for (int i = 0; i < dataN; ++i)
		{
			for (int d = 0; d < samples; ++d)
			{
				for (int j = 0; j < stride; ++j)
				{
					(*outfirst)[i * stride + j] = iir_filter_biquad((*inp)[(i * samples + d) * stride + j], coeffdec, sections, *hist + 2 * j * sections);
				}
			}
		}
	}
	write_results(*outfirst, MAX_WRITE_RESULT, "reference_filter.txt", 1);
	return 1;
}

// ---------------------------------------------------------------------------------------------
// test_setup_demod_stride
_Bool test_setup_demod_stride(double* inp, double* out, double* outfirst, double* hist, avx_stride_struct** avxstridehist)
{
	if ((inp == NULL) || (out == NULL) || (outfirst == NULL) || (hist == NULL) || (avxstridehist == NULL)) {
		return 0;
	}
	// ---------------------------------------------------------------------------------------------
	// zero history
	printf("\ngenerate reference data (demodulation_decimation_stride8_section3_std)\n");
	memset(outfirst, 0, 2 * dataN * stride * sizeof(double));
	memset(hist, 0, 4 * stride * sections * sizeof(double));

	// ---------------------------------------------------------------------------------------------
	// set up rotation parameters
	*avxstridehist = (avx_stride_struct*)_mm_malloc(sizeof(avx_stride_struct), alignment);
	if (*avxstridehist == NULL) {
		return 0;
	}
	avx_stride_struct_init(*avxstridehist, 1.0);
	avx_stride_struct_set_frequency(*avxstridehist, demodfreq);

	// ---------------------------------------------------------------------------------------------
	// fill reference data by using standard demodulation/biquad
	for (int k = 0; k < repeattest; ++k)
	{
		for (int i = 0; i < dataN; ++i)
		{
			for (int d = 0; d < samples; ++d)
			{
				demodulation_decimation_stride8_section3_std(inp + (i * samples + d) * stride, demodfreq, outfirst + 2 * (i * stride), coeffdec, *avxstridehist, stride);
			}
		}
	}
	// write results
	write_demod_results(out, MAX_WRITE_RESULT, "reference_demodulation_1.txt", samples);
	
	return 1;
}

// ---------------------------------------------------------------------------------------------
// test_setup_demod_rotation
_Bool test_setup_demod_rotation(double* inp, double* out, double* outfirst, double* hist, avx_rotation_struct** avxrothist)
{
	if ((inp == NULL) || (out == NULL) || (outfirst == NULL) || (hist == NULL) || (avxrothist == NULL)) {
		return 0;
	}
	// ---------------------------------------------------------------------------------------------
	// zero history
	printf("\ngenerate reference data (demodulation_decimation_rotation8_section3_std)\n");
	memset(outfirst, 0, 2 * dataN * stride * sizeof(double));
	memset(hist, 0, 4 * stride * sections * sizeof(double));

	// ---------------------------------------------------------------------------------------------
	// set up rotation parameters
	*avxrothist = (avx_rotation_struct*)_mm_malloc(sizeof(avx_rotation_struct), alignment);
	if (*avxrothist == NULL) {
		return 0;
	}
	avx_rotation_struct_init(*avxrothist, 1.0);
	for (unsigned int j = 0; j < stride; ++j)
	{
		demodfreqarr[j] = 0.01 * (25.0 + (j - 16.0) * 0.1);
		avx_rotation_struct_set_frequency(*avxrothist, demodfreqarr[j], j);
	}

	// ---------------------------------------------------------------------------------------------
	// fill reference data by using standard demodulation/biquad
	for (int k = 0; k < repeattest; ++k)
	{
		for (int i = 0; i < dataN; ++i)
		{
			for (int d = 0; d < samples; ++d)
			{
				demodulation_decimation_rotation8_section3_std(inp[(i * samples + d) * stride], demodfreqarr, outfirst + 2 * (i * stride), coeffdec, *avxrothist, stride);
			}
		}
	}
	// write results
	write_demod_results(out, MAX_WRITE_RESULT, "reference_demodulation_2.txt", samples);

	return 1;
}

// ---------------------------------------------------------------------------------------------
// test_cleanup
void test_cleanup(double* inp, double* out, double* outfirst, double* hist)
{
	_mm_free(inp);
	_mm_free(out);
	_mm_free(outfirst);
	_mm_free(hist);
}


// ---------------------------------------------------------------------------------------------
// test_biquad2_sse3
double test_biquad2_sse3(double* inp, double* out, double* outfirst, double* hist)
{
	assert(inp && out && outfirst && hist);
	double diff = 0.0;

	// ---------------------------------------------------------------------------------------------
	// zero history
#if USE_SSE3
	printf("\nbiquad2_sse3:\n");
	memset(out, 0, dataN * stride * sizeof(double));
	memset(hist, 0, 2 * stride * sections * sizeof(double));

	// run biquad2_sse3
	for (int k = 0; k < repeattest; ++k) {
		for (int i = 0; i < dataN; ++i)
		{
			for (int d = 0; d < samples; ++d)
			{
				for (int j = 0; j <= stride - 2; j += 2)
				{
					biquad2_sse3(inp + (i * samples + d) * stride + j, out + i * stride + j, coeffdec2, hist + 2 * j * sections, sections);
				}
			}
		}
	}
	diff = check_result(out, outfirst, dataN * stride);
	printf("Biggest deviation %g\n", diff);
	write_results(out, MAX_WRITE_RESULT, "test_biquad2_sse3.txt", 1);
#endif

	return diff;
}

// ---------------------------------------------------------------------------------------------
// test_biquad4_avx2
double test_biquad4_avx2(double* inp, double* out, double* outfirst, double* hist)
{
	assert(inp && out && outfirst && hist);
	double diff = 0.0;

	// ---------------------------------------------------------------------------------------------
	// zero history
#if USE_AVX2
	printf("\nbiquad4_avx2:\n");
	memset(out, 0, dataN * stride * sizeof(double));
	memset(hist, 0, 2 * stride * sections * sizeof(double));

	// run biquad4_avx2
	for (int k = 0; k < repeattest; ++k) {
		for (int i = 0; i < dataN; ++i)
		{
			for (int d = 0; d < samples; ++d)
			{
				for (int j = 0; j <= stride - 4; j += 4)
				{
					biquad4_avx2(inp + (i * samples + d) * stride + j, out + i * stride + j, coeffdec4, hist + 2 * j * sections, sections);
				}
			}
		}
	}
	diff = check_result(out, outfirst, dataN * stride);
	printf("Biggest deviation %g\n", diff);
	write_results(out, MAX_WRITE_RESULT, "testbiquad4_avx2.txt", 1);
#endif

	return diff;
}

// ---------------------------------------------------------------------------------------------
// test_biquad8_avx512
double test_biquad8_avx512(double* inp, double* out, double* outfirst, double* hist)
{
	assert(inp && out && outfirst && hist);
	double diff = 0.0;

	// ---------------------------------------------------------------------------------------------
	// zero history
#if USE_AVX512
	printf("\nbiquad8_avx512:\n");
	memset(out, 0, dataN * stride * sizeof(double));
	memset(hist, 0, 2 * stride * sections * sizeof(double));

	// run biquad8_avx512
	for (int k = 0; k < repeattest; ++k) {
		for (int i = 0; i < dataN; ++i)
		{
			for (int d = 0; d < samples; ++d)
			{
				for (int j = 0; j <= stride - 8; j += 8)
				{
					biquad8_avx512(inp + (i * samples + d) * stride + j, out + i * stride + j, coeffdec8, hist + 2 * j * sections, sections);
				}
			}
		}
	}
	diff = check_result(out, outfirst, dataN * stride);
	printf("Biggest deviation %g\n", diff);
	write_results(out, MAX_WRITE_RESULT, "test_biquad8_avx512.txt", 1);
#endif

	return diff;
}

// ---------------------------------------------------------------------------------------------
// test_biquad_stride2_std
double test_biquad_stride2_std(double* inp, double* out, double* outfirst, double* hist)
{
	assert(inp && out && outfirst && hist);
	double diff = 0.0;

	// ---------------------------------------------------------------------------------------------
	// zero history
	printf("\nbiquad_stride2_std:\n");
	memset(out, 0, dataN * stride * sizeof(double));
	memset(hist, 0, 2 * stride * sections * sizeof(double));

	// run biquad_stride2_std
	for (int k = 0; k < repeattest; ++k) {
		for (int i = 0; i < dataN; ++i)
		{
			biquad_stride2_std(inp + i * stride * samples, out + i * stride, coeffdec, hist, sections, stride, samples, 0);
		}
	}
	diff = check_result(out, outfirst, dataN * stride);
	printf("Biggest deviation %g\n", diff);
	write_results(out, MAX_WRITE_RESULT, "test_biquad_stride2_std.txt", 1);

	return diff;
}

// ---------------------------------------------------------------------------------------------
// test_biquad_stride2_section3_sse3
double test_biquad_stride2_section3_sse3(double* inp, double* out, double* outfirst, double* hist)
{
	assert(inp && out && outfirst && hist);
	double diff = 0.0;

	// ---------------------------------------------------------------------------------------------
	// zero history
#if USE_SSE3
	printf("\nbiquad_stride2_section3_sse3:\n");
	memset(out, 0, dataN * stride * sizeof(double));
	memset(hist, 0, 2 * stride * sections * sizeof(double));

	// run biquad_stride2_section3_sse3
	for (int k = 0; k < repeattest; ++k) {
		for (int i = 0; i < dataN; ++i)
		{
			biquad_stride2_section3_sse3(inp + i * stride * samples, out + i * stride, coeffdec, hist, stride, samples, 0);
		}
	}
	diff = check_result(out, outfirst, dataN * stride);
	printf("Biggest deviation %g\n", diff);
	write_results(out, MAX_WRITE_RESULT, "test_biquad_stride2_section3_sse3.txt", 1);
#endif

	return diff;
}

// ---------------------------------------------------------------------------------------------
// test_biquad_stride8_section3_sse3
double test_biquad_stride8_section3_sse3(double* inp, double* out, double* outfirst, double* hist)
{
	assert(inp && out && outfirst && hist);
	double diff = 0.0;

	// ---------------------------------------------------------------------------------------------
	// zero history
#if USE_SSE3
	printf("\nbiquad_stride8_section3_sse3:\n");
	memset(out, 0, dataN * stride * sizeof(double));
	memset(hist, 0, 2 * stride * sections * sizeof(double));

	// run 
	for (int k = 0; k < repeattest; ++k) {
		for (int i = 0; i < dataN; ++i)
		{
			biquad_stride8_section3_sse3(inp + i * stride * samples, out + i * stride, coeffdec, hist, stride, samples, 0);
		}
	}
	diff = check_result(out, outfirst, dataN * stride);
	printf("Biggest deviation %g\n", diff);
	write_results(out, MAX_WRITE_RESULT, "test_biquad_stride8_section3_sse3.txt", 1);
#endif

	return diff;
}

// ---------------------------------------------------------------------------------------------
// test_biquad_stride4_section3_avx2
double test_biquad_stride4_section3_avx2(double* inp, double* out, double* outfirst, double* hist)
{
	assert(inp && out && outfirst && hist);
	double diff = 0.0;

	// ---------------------------------------------------------------------------------------------
	// zero history
#if USE_AVX2
	printf("\nbiquad_stride4_section3_avx2:\n");
	memset(out, 0, dataN * stride * sizeof(double));
	memset(hist, 0, 2 * stride * sections * sizeof(double));

	// run biquad_stride4_section3_avx2
	for (int k = 0; k < repeattest; ++k) {
		for (int i = 0; i < dataN; ++i)
		{
			biquad_stride4_section3_avx2(inp + i * stride * samples, out + i * stride, coeffdec, hist, stride, samples, 0);
		}
	}
	diff = check_result(out, outfirst, dataN * stride);
	printf("Biggest deviation %g\n", diff);
	write_results(out, MAX_WRITE_RESULT, "test_biquad_stride4_section3_avx2.txt", 1);
#endif

	return diff;
}

// ---------------------------------------------------------------------------------------------
// test_biquad_stride16_section3_avx2
double test_biquad_stride16_section3_avx2(double* inp, double* out, double* outfirst, double* hist)
{
	assert(inp && out && outfirst && hist);
	double diff = 0.0;

	// ---------------------------------------------------------------------------------------------
	// zero history
#if USE_AVX2
	printf("\nbiquad_stride16_section3_avx2:\n");
	memset(out, 0, dataN * stride * sizeof(double));
	memset(hist, 0, 2 * stride * sections * sizeof(double));

	// run biquad_stride16_section3_avx2
	for (int k = 0; k < repeattest; ++k) {
		for (int i = 0; i < dataN; ++i)
		{
			biquad_stride16_section3_avx2(inp + i * stride * samples, out + i * stride, coeffdec, hist, stride, samples, 0);
		}
	}
	diff = check_result(out, outfirst, dataN * stride);
	printf("Biggest deviation %g\n", diff);
	write_results(out, MAX_WRITE_RESULT, "test_biquad_stride16_section3_avx2.txt", 1);
#endif

	return diff;
}

// ---------------------------------------------------------------------------------------------
// test_biquad_stride16_section3_avx512
double test_biquad_stride16_section3_avx512(double* inp, double* out, double* outfirst, double* hist)
{
	assert(inp && out && outfirst && hist);
	double diff = 0.0;

	// ---------------------------------------------------------------------------------------------
	// zero history
#if USE_AVX512
	printf("\nbiquad_stride16_section3_avx512:\n");
	memset(out, 0, dataN * stride * sizeof(double));
	memset(hist, 0, 2 * stride * sections * sizeof(double));

	// run biquad_stride16_section3_avx512
	for (int k = 0; k < repeattest; ++k) {
		for (int i = 0; i < dataN; ++i)
		{
			biquad_stride16_section3_avx512(inp + i * stride * samples, out + i * stride, coeffdec, hist, stride, samples, 0);
		}
	}
	diff = check_result(out, outfirst, dataN * stride);
	printf("Biggest deviation %g\n", diff);
	write_results(out, MAX_WRITE_RESULT, "test_biquad_stride16_section3_avx512.txt", 1);
#endif

	return diff;
}

// ---------------------------------------------------------------------------------------------
// test_demodulation_decimation_stride8_section3_sse3
double test_demodulation_decimation_stride8_section3_sse3(double* inp, double* out, double* outfirst, double* hist, avx_stride_struct* avxstridehist)
{
	assert(inp && out && outfirst && hist && avxstridehist);
	double diff = 0.0;

	// ---------------------------------------------------------------------------------------------
	// zero history
#if USE_SSE3
	printf("\ndemodulation_decimation_stride8_section3_sse3:\n");
	memset(out, 0, 2 * dataN * stride * sizeof(double));
	memset(hist, 0, 4 * stride * sections * sizeof(double));
	// init cordic
	avx_stride_struct_init(avxstridehist, 1.0);
	avx_stride_struct_set_frequency(avxstridehist, demodfreq);

	// run demodulation_decimation_stride8_section3_sse3
	for (int k = 0; k < repeattest; ++k)
	{
		for (int i = 0; i < dataN; ++i)
		{
			for (int d = 0; d < samples; ++d)
			{
				demodulation_decimation_stride8_section3_sse3(inp + (i * samples + d) * stride, demodfreq, out + 2 * i * stride, coeffdec, avxstridehist, stride);
			}
		}
	}
	diff = check_result(out, outfirst, 2 * dataN * stride);
	printf("Biggest deviation %g\n", diff);
	write_results(out, MAX_WRITE_RESULT, "test_demodulation_decimation_stride8_section3_sse3.txt", 1);
#endif

	return diff;
}

// ---------------------------------------------------------------------------------------------
// test_demodulation_decimation_stride8_section3_avx2
double test_demodulation_decimation_stride8_section3_avx2(double* inp, double* out, double* outfirst, double* hist, avx_stride_struct* avxstridehist)
{
	assert(inp && out && outfirst && hist && avxstridehist);
	double diff = 0.0;

	// ---------------------------------------------------------------------------------------------
	// zero history
#if USE_AVX2
	printf("\ndemodulation_decimation_stride8_section3_avx2:\n");
	memset(out, 0, 2 * dataN * stride * sizeof(double));
	memset(hist, 0, 4 * stride * sections * sizeof(double));
	// init cordic
	avx_stride_struct_init(avxstridehist, 1.0);
	avx_stride_struct_set_frequency(avxstridehist, demodfreq);

	// run 
	for (int k = 0; k < repeattest; ++k)
	{
		for (int i = 0; i < dataN; ++i)
		{
			for (int d = 0; d < samples; ++d)
			{
				demodulation_decimation_stride8_section3_avx2(inp + (i * samples + d) * stride, demodfreq, out + 2 * i * stride, coeffdec, avxstridehist, stride);
			}
		}
	}
	diff = check_result(out, outfirst, 2 * dataN * stride);
	printf("Biggest deviation %g\n", diff);
	write_results(out, MAX_WRITE_RESULT, "test_demodulation_decimation_stride8_section3_avx2.txt", 1);
#endif

	return diff;
}

// ---------------------------------------------------------------------------------------------
// test_demodulation_decimation_stride8_section3_avx512
double test_demodulation_decimation_stride8_section3_avx512(double* inp, double* out, double* outfirst, double* hist, avx_stride_struct* avxstridehist)
{
	assert(inp && out && outfirst && hist && avxstridehist);
	double diff = 0.0;

	// ---------------------------------------------------------------------------------------------
	// zero history
#if USE_AVX512
	printf("\ndemodulation_decimation_stride8_section3_avx512:\n");
	memset(out, 0, 2 * dataN * stride * sizeof(double));
	memset(hist, 0, 4 * stride * sections * sizeof(double));
	// init cordic
	avx_stride_struct_init(avxstridehist, 1.0);
	avx_stride_struct_set_frequency(avxstridehist, demodfreq);

	// run 
	for (int k = 0; k < repeattest; ++k)
	{
		for (int i = 0; i < dataN; ++i)
		{
			for (int d = 0; d < samples; ++d)
			{
				demodulation_decimation_stride8_section3_avx512(inp + (i * samples + d) * stride, demodfreq, out + 2 * i * stride, coeffdec, avxstridehist, stride);
			}
		}
	}
	diff = check_result(out, outfirst, 2 * dataN * stride);
	printf("Biggest deviation %g\n", diff);
	write_results(out, MAX_WRITE_RESULT, "test_demodulation_decimation_stride8_section3_avx512.txt", 1);
#endif

	return diff;
}

// ---------------------------------------------------------------------------------------------
// test_demodulation_decimation_rotation8_section3_sse3
double test_demodulation_decimation_rotation8_section3_sse3(double* inp, double* out, double* outfirst, double* hist, avx_rotation_struct* avxrothist)
{
	assert(inp && out && outfirst && hist && avxrothist);
	double diff = 0.0;

	// ---------------------------------------------------------------------------------------------
	// zero history
#if USE_SSE3
	printf("\ndemodulation_decimation_rotation8_section3_sse3:\n");
	memset(out, 0, 2 * dataN * stride * sizeof(double));
	memset(hist, 0, 4 * stride * sections * sizeof(double));
	// init cordic
	avx_rotation_struct_init(avxrothist, 1.0);
	for (unsigned int j = 0; j < stride; ++j)
	{
		avx_rotation_struct_set_frequency(avxrothist, demodfreqarr[j], j);
	}

	// run demodulation_decimation_rotation8_section3_sse3
	for (int k = 0; k < repeattest; ++k)
	{
		for (int i = 0; i < dataN; ++i)
		{
			for (int d = 0; d < samples; ++d)
			{
				demodulation_decimation_rotation8_section3_sse3(inp[(i * samples + d) * stride], demodfreqarr, out + 2 * (i * stride), coeffdec, avxrothist, stride);
			}
		}
	}
	diff = check_result(out, outfirst, 2 * dataN * stride);
	printf("Biggest deviation %g\n", diff);
	write_results(out, MAX_WRITE_RESULT, "test_demodulation_decimation_rotation8_section3_sse3.txt", 1);
#endif

	return diff;
}

// ---------------------------------------------------------------------------------------------
// test_demodulation_decimation_rotation8_section3_avx2
double test_demodulation_decimation_rotation8_section3_avx2(double* inp, double* out, double* outfirst, double* hist, avx_rotation_struct* avxrothist)
{
	assert(inp && out && outfirst && hist && avxrothist);
	double diff = 0.0;

	// ---------------------------------------------------------------------------------------------
	// zero history
#if USE_AVX2
	printf("\ndemodulation_decimation_rotation8_section3_avx2:\n");
	memset(out, 0, 2 * dataN * stride * sizeof(double));
	memset(hist, 0, 4 * stride * sections * sizeof(double));
	// init cordic
	avx_rotation_struct_init(avxrothist, 1.0);
	for (unsigned int j = 0; j < stride; ++j)
	{
		avx_rotation_struct_set_frequency(avxrothist, demodfreqarr[j], j);
	}

	// run 
	for (int k = 0; k < repeattest; ++k)
	{
		for (int i = 0; i < dataN; ++i)
		{
			for (int d = 0; d < samples; ++d)
			{
				demodulation_decimation_rotation8_section3_avx2(inp[(i * samples + d) * stride], demodfreqarr, out + 2 * (i * stride), coeffdec, avxrothist, stride);
			}
		}
	}
	diff = check_result(out, outfirst, 2 * dataN * stride);
	printf("Biggest deviation %g\n", diff);
	write_results(out, MAX_WRITE_RESULT, "test_demodulation_decimation_rotation8_section3_avx2.txt", 1);
#endif

	return diff;
}

// ---------------------------------------------------------------------------------------------
// test_demodulation_decimation_rotation8_section3_avx512
double test_demodulation_decimation_rotation8_section3_avx512(double* inp, double* out, double* outfirst, double* hist, avx_rotation_struct* avxrothist)
{
	assert(inp && out && outfirst && hist && avxrothist);
	double diff = 0.0;

	// ---------------------------------------------------------------------------------------------
	// zero history
#if USE_AVX512
	printf("\ndemodulation_decimation_rotation8_section3_avx512:\n");
	memset(out, 0, 2 * dataN * stride * sizeof(double));
	memset(hist, 0, 4 * stride * sections * sizeof(double));
	// init cordic
	avx_rotation_struct_init(avxrothist, 1.0);
	for (unsigned int j = 0; j < stride; ++j)
	{
		avx_rotation_struct_set_frequency(avxrothist, demodfreqarr[j], j);
	}

	// run 
	for (int k = 0; k < repeattest; ++k)
	{
		for (int i = 0; i < dataN; ++i)
		{
			for (int d = 0; d < samples; ++d)
			{
				demodulation_decimation_rotation8_section3_avx512(inp[(i * samples + d) * stride], demodfreqarr, out + 2 * (i * stride), coeffdec, avxrothist, stride);
			}
		}
	}
	diff = check_result(out, outfirst, 2 * dataN * stride);
	printf("Biggest deviation %g\n", diff);
	write_results(out, MAX_WRITE_RESULT, "test_demodulation_decimation_rotation8_section3_avx512.txt", 1);
#endif

	return diff;
}


// ---------------------------------------------------------------------------------------------
// main
int main()
{
	double* inp = NULL;
	double* out = NULL;
	double* outfirst = NULL;
	double* hist = NULL;
	avx_stride_struct* avxstridehist = NULL;
	avx_rotation_struct* avxrothist = NULL;
	double diff = 0.0;
	
	printf("BIQUAD UNIT TEST\n");

	// ---------------------------------------------------------------------------------------------
	// test setup
	if (!test_setup(&inp, &out, &outfirst, &hist)) return 1;

	// ---------------------------------------------------------------------------------------------
	// run filter tests
	diff = test_biquad2_sse3(inp, out, outfirst, hist);
	assert(diff < tolerance);

	diff = test_biquad4_avx2(inp, out, outfirst, hist);
	assert(diff < tolerance);

	diff = test_biquad8_avx512(inp, out, outfirst, hist);
	assert(diff < tolerance);

	diff = test_biquad_stride2_std(inp, out, outfirst, hist);
	assert(diff < tolerance);

	diff = test_biquad_stride2_section3_sse3(inp, out, outfirst, hist);
	assert(diff < tolerance);

	diff = test_biquad_stride8_section3_sse3(inp, out, outfirst, hist);
	assert(diff < tolerance);

	diff = test_biquad_stride4_section3_avx2(inp, out, outfirst, hist);
	assert(diff < tolerance);

	diff = test_biquad_stride16_section3_avx2(inp, out, outfirst, hist);
	assert(diff < tolerance);

	diff = test_biquad_stride16_section3_avx512(inp, out, outfirst, hist);
	assert(diff < tolerance);

	// ---------------------------------------------------------------------------------------------
	// setup demodulation tests for stride
	if (!test_setup_demod_stride(inp, out, outfirst, hist, &avxstridehist)) return 1;

	// ---------------------------------------------------------------------------------------------
	// run demodulation stride tests
	diff = test_demodulation_decimation_stride8_section3_sse3(inp, out, outfirst, hist, avxstridehist);
	assert(diff < tolerance);

	diff = test_demodulation_decimation_stride8_section3_avx2(inp, out, outfirst, hist, avxstridehist);
	assert(diff < tolerance);
	
	diff = test_demodulation_decimation_stride8_section3_avx512(inp, out, outfirst, hist, avxstridehist);
	assert(diff < tolerance);

	// ---------------------------------------------------------------------------------------------
	// setup demodulation tests for rotation
	if (!test_setup_demod_rotation(inp, out, outfirst, hist, &avxrothist)) return 1;

	// ---------------------------------------------------------------------------------------------
	// run demodulation rotation tests
	diff = test_demodulation_decimation_rotation8_section3_sse3(inp, out, outfirst, hist, avxrothist);
	assert(diff < tolerance);

	diff = test_demodulation_decimation_rotation8_section3_avx2(inp, out, outfirst, hist, avxrothist);
	assert(diff < tolerance);

	diff = test_demodulation_decimation_rotation8_section3_avx512(inp, out, outfirst, hist, avxrothist);
	assert(diff < tolerance);

	return 0;
}
