// biquadAVX.cpp : This file contains the 'main' function. Program execution begins and ends there.
//
#pragma warning(disable : 4514 4701 4710 4711 4820 5045)

#define _CRT_SECURE_NO_WARNINGS
#define CHECK_RESULT
#define MAX_WRITE_RESULT 1000
#include <chrono>
#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include <string.h>
#include <stdalign.h>
#include "Biquad.h"
#include "BiquadTest.h"
#include "coeff_traits.h"
#include "mxcsr.h"
#include "instructionset.h"

const double twopi = 6.2831853071795864769;

const int dataN = 1000;
const int stride = 32;
const int samples = 8;
const int alignment = 64;
const int sections = 3;

#ifdef _DEBUG
const int repeattest = 1;
#else
const int repeattest = 10000;
#endif

// Coefficient array for 3 second order section (6th order IIR filter)
using coeff_traits3 = coeff_traits<3>;
using coeff_order_6 = coeff_traits3::coeff_array;
using coeff_order_6_wide_2 = coeff_traits3::coeff_wide_array<2>;
using coeff_order_6_wide_4 = coeff_traits3::coeff_wide_array<4>;
using coeff_order_6_wide_8 = coeff_traits3::coeff_wide_array<8>;


//static const double coeffdec[13] = { 0.00162185538923,
//							  0.73342532779703, -0.02862365091314, 1.44110125961504,  1.67905228090487,
//							  0.77657563380963, -0.08304311675394, 0.18851328163424,  0.32889453107067,
//							  0.83213081484618, -0.12573495191273, 0.0940911979108501,0.13622543115194 };

// cheby2("LowPass",6,60,8000)
// rate 524288 Hz
//static const coeff_order_6 chebydirect = { 9.410818796785352805325808e-04,
//							 -1.9752147772326503,  0.9777400275124176, -1.9901574406868523,  1.0000000000000000,
//							 -1.9296605059414182,  0.9324647855643438, -1.9816725895662000,  1.0000000000000000,
//							 -1.8920342775852999,  0.8952193075414041, -1.8671374119651403,  0.9999999999999999 };

// cheby2("LowPass",6,60,32000)
// rate 524288 Hz
alignas(8) static coeff_order_6 chebydirect = { 1.159664984828962224633342e-03,
							 -1.8740625376565518,  0.9136862817567704, -1.8446997352645189,  1.0000000000000000,
							 -1.7133654266738769,  0.7545999972582751, -1.7196326892723282,  1.0000000000000000,
							 -1.5936647899894101,  0.6381719943234752, -0.5598252720330773,  1.0000000000000000 };

alignas(16) static coeff_order_6_wide_2 chebydirect2{ coeff_traits3::wide(chebydirect, chebydirect) };
alignas(32) static coeff_order_6_wide_4 chebydirect4{ coeff_traits3::wide(chebydirect, chebydirect, chebydirect, chebydirect) };
alignas(64) static coeff_order_6_wide_8 chebydirect8{ coeff_traits3::wide(chebydirect, chebydirect, chebydirect, chebydirect, chebydirect, chebydirect, chebydirect, chebydirect) };

alignas(8) static coeff_order_6 chebybiquad = coeff_traits3::biquad_shuffle(chebydirect);
alignas(16) static coeff_order_6_wide_2 chebybiquad2{ coeff_traits3::wide(chebybiquad, chebybiquad) };
alignas(32) static coeff_order_6_wide_4 chebybiquad4{ coeff_traits3::wide(chebybiquad, chebybiquad, chebybiquad, chebybiquad)};
alignas(64) static coeff_order_6_wide_8 chebybiquad8{ coeff_traits3::wide(chebybiquad, chebybiquad, chebybiquad, chebybiquad, chebybiquad, chebybiquad, chebybiquad, chebybiquad) };

#ifdef LIGO_BIQUAD
const double* coeff = chebybiquad.data();
const double* coeff2 = chebybiquad2.data();
const double* coeff4 = chebybiquad4.data();
const double* coeff8 = chebybiquad8.data();
#else
const double* coeff = chebydirect.data();
const double* coeff2 = chebydirect2.data();
const double* coeff4 = chebydirect4.data();
const double* coeff8 = chebydirect8.data();
#endif

#ifdef __GNUC__
static const char* const dirpath = R"(~/)";
#else
static const char* const dirpath = R"(C:\SlowControls\EPICS\Utilities\BiquadAVX\)";
#endif

static
void write_results (double* out, int n, const char* filename, int offset = samples)
{
	int max = n;
	if (max > dataN) max = dataN;
	if (max <= 0) return;
	if (offset < 0) return;
    if (!filename) return;

	char filen[1024];
#ifdef __GNUC__
	if ((strlen(filename) >= 1) && ((filename[0] == '~') || (filename[0] == '/'))) {
#else
	if ((strlen(filename) >= 2) && (filename[1] == ':')) {
#endif
		strncpy(filen, filename, sizeof(filen));
		filen[sizeof(filen) - 1] = 0;
	}
	else {
		strncpy(filen, dirpath, sizeof(filen));
		filen[sizeof(filen) - 1] = 0;
		size_t len = strlen(filen);
		if (len < sizeof(filen)) strncpy(filen + len, filename, sizeof(filen) - len);
		filen[sizeof(filen) - 1] = 0;
	}

	FILE* fout = fopen(filename, "w");
	if (!fout) return;

	for (int i = 0; i < max; ++i)
	{
		fprintf(fout, "%4i %12.9f %12.9f %12.9f %12.9f %12.9f %12.9f %12.9f %12.9f\n", i * offset,
				out[i * stride + 0], out[i * stride + 4], out[i * stride + 8], out[i * stride + 12],
				out[i * stride + 16], out[i * stride + 20], out[i * stride + 24], out[i * stride + 28]);
	}
	fclose(fout);
}

#ifdef CHECK_RESULT
static
double check_result(const double* inp1, const double* inp2, int n)
{
	double max = 0;
	for (int i = 0; i < n; ++i)
	{
		double diff = abs (inp1[i] - inp2[i]);
		if (isnan(inp1[i]) || isnan(inp2[i])) diff = 1E9;
		if (diff > max) max = diff;
	}
	return max;
}
#endif

int main()
{
	using clock = std::chrono::system_clock;
	using sec = std::chrono::duration<double>;

	printf("BIQUAD EXECUTATION TIME COMPARISON\n");
	InstructionSet::PRINTAVX();
	fz_daz();
	print_mxcsr();

	double* inp = (double*)_mm_malloc(dataN * stride * samples * sizeof(double), alignment);
	double* out = (double*) _mm_malloc(dataN * stride * sizeof(double), alignment);
	double* hist = (double*)_mm_malloc(2 * stride * sections * sizeof(double), alignment);

	// ---------------------------------------------------------------------------------------------
	// initialize input
	for (int i = 0; i < dataN * samples; ++i)
	{
		for (int j = 0; j < stride; ++j) 
		{
			if (inp) inp[i * stride + j] = sin(twopi * i * 0.01 * (j+1));
		}
	}
	write_results(inp, MAX_WRITE_RESULT, R"(file_inp.txt)", 1);

	// ---------------------------------------------------------------------------------------------
	// zero history
	printf("\niir_filter_biquad:\n");
	if (out) memset(out, 0, dataN * stride * sizeof(double));
	if (hist) memset(hist, 0, 2 * stride * sections * sizeof(double));

	// time iir_filter_biquad
	auto before = clock::now();
	for (int k = 0; k < repeattest; ++k) {
		for (int i = 0; i < dataN; ++i)
		{
			for (int d = 0; d < samples; ++d) 
			{
				for (int j = 0; j < stride; ++j)
				{
					out[i * stride + j] = inline_iir_filter_biquad(inp[(i * samples + d) * stride + j], chebybiquad.data(), sections, hist + 2 * j * sections);
				}
			}
		}
	}
	sec duration0 = clock::now() - before;
	printf("It took %g sec\n", duration0.count());

	// write results
	write_results(out, MAX_WRITE_RESULT, R"(file_iir_filter_biquad.txt)");
#ifdef CHECK_RESULT
	double* outfirst = out;
	out = (double*)_mm_malloc(dataN * stride * sizeof(double), alignment);
#endif

	// ---------------------------------------------------------------------------------------------
	// zero history
	printf("\niir_filter_biquad_sse3:\n");
	if (out) memset(out, 0, dataN * stride * sizeof(double));
	if (hist) memset(hist, 0, 2 * stride * sections * sizeof(double));

	// time biquad_stride2_std
	for (int k = 0; k < repeattest; ++k) {
		for (int i = 0; i < dataN; ++i)
		{
			for (int d = 0; d < samples; ++d)
			{
				for (int j = 0; j < stride; ++j)
				{
					out[i * stride + j] = iir_filter_biquad_sse3(inp[(i * samples + d) * stride + j], chebybiquad.data(), sections, hist + 2 * j * sections);
				}
			}
		}
	}
	sec duration_biquad_sse = clock::now() - before;
	printf("It took %g sec\n", duration_biquad_sse.count());
	printf("It was %gx faster\n", duration0.count() / duration_biquad_sse.count());

#ifdef CHECK_RESULT
	double diff = check_result(out, outfirst, dataN * stride);
	printf("Biggest deviation %g\n", diff);
#endif
	write_results(out, MAX_WRITE_RESULT, R"(file_biquad_sse.txt)", 1);

	// ---------------------------------------------------------------------------------------------
	// zero history
	printf("\nbiquad_std:\n");
	if (out) memset(out, 0, dataN * stride * sizeof(double));
	if (hist) memset(hist, 0, 2 * stride * sections * sizeof(double));

	// time biquad 
	before = clock::now();
	for (int k = 0; k < repeattest; ++k) {
		for (int i = 0; i < dataN; ++i)
		{
			for (int d = 0; d < samples; ++d)
			{
				for (int j = 0; j < stride; ++j)
				{
					biquad_std(inp + (i * samples + d) * stride + j, out + i * stride + j, coeff, hist + 2 * j * sections, sections);
				}
			}
		}
	}
	sec duration_biquad = clock::now() - before;
	printf("It took %g sec\n", duration_biquad.count());
	printf("It was %gx faster\n", duration0.count() / duration_biquad.count());

#ifdef CHECK_RESULT
	diff = check_result(out, outfirst, dataN * stride);
	printf("Biggest deviation %g\n", diff);
#endif
	// write results
	write_results(out, MAX_WRITE_RESULT, R"(file_biquad_std.txt)");

	// ---------------------------------------------------------------------------------------------
	// zero history
	printf("\nbiquad2_sse3:\n");
	if (out) memset(out, 0, dataN * stride * sizeof(double));
	if (hist) memset(hist, 0, 2 * stride * sections * sizeof(double));

	// time biquad2_sse3
	before = clock::now();
	for (int k = 0; k < repeattest; ++k) {
		for (int i = 0; i < dataN; ++i)
		{
			for (int d = 0; d < samples; ++d)
			{
				for (int j = 0; j <= stride - 2; j += 2)
				{
					biquad2_sse3(inp + (i * samples + d) * stride + j, out + i * stride + j, coeff2, hist + 2 * j * sections, sections);
				}
			}
		}
	}
	sec duration_biquadSSE3 = clock::now() - before;
	printf("It took %g sec\n", duration_biquadSSE3.count());
	printf("It was %gx faster\n", duration0.count() / duration_biquadSSE3.count());

#ifdef CHECK_RESULT
	diff = check_result(out, outfirst, dataN * stride);
	printf("Biggest deviation %g\n", diff);
#endif
	// write results
	write_results(out, MAX_WRITE_RESULT, R"(file_biquad2_sse3.txt)");

	// ---------------------------------------------------------------------------------------------
	// zero history
	printf("\nbiquad4_avx2:\n");
	if (out) memset(out, 0, dataN * stride * sizeof(double));
	if (hist) memset(hist, 0, 2 * stride * sections * sizeof(double));

	// time biquad4_avx2
	before = clock::now();
	for (int k = 0; k < repeattest; ++k) {
		for (int i = 0; i < dataN; ++i)
		{
			for (int d = 0; d < samples; ++d)
			{
				for (int j = 0; j <= stride - 4; j += 4)
				{
					biquad4_avx2(inp + (i * samples + d) * stride + j, out + i * stride + j, coeff4, hist + 2 * j * sections, sections);
				}
			}
		}
	}
	sec duration_biquadAVX2 = clock::now() - before;
	printf("It took %g sec\n", duration_biquadAVX2.count());
	printf("It was %gx faster\n", duration0.count() / duration_biquadAVX2.count());

#ifdef CHECK_RESULT
	diff = check_result(out, outfirst, dataN * stride);
	printf("Biggest deviation %g\n", diff);
#endif
	// write results
	write_results(out, MAX_WRITE_RESULT, R"(file_biquad4_avx2.txt)");

	// ---------------------------------------------------------------------------------------------
	// zero history
	printf("\nbiquad8_avx2:\n");
	if (out) memset(out, 0, dataN * stride * sizeof(double));
	if (hist) memset(hist, 0, 2 * stride * sections * sizeof(double));

	// time biquad8_avx2
	before = clock::now();
	for (int k = 0; k < repeattest; ++k) 
	{
		for (int i = 0; i < dataN; ++i)
		{
			for (int d = 0; d < samples; ++d)
			{
				for (int j = 0; j <= stride - 8; j += 8)
				{
					biquad8_avx2(inp + (i * samples + d) * stride + j, out + i * stride + j, coeff8, hist + 2 * j * sections, sections);
				}
			}
		}
	}
	sec duration_biquadAVX2double = clock::now() - before;
	printf("It took %g sec\n", duration_biquadAVX2double.count());
	printf("It was %gx faster\n", duration0.count() / duration_biquadAVX2double.count());

#ifdef CHECK_RESULT
	diff = check_result(out, outfirst, dataN * stride);
	printf("Biggest deviation %g\n", diff);
#endif
	// write results
	write_results(out, MAX_WRITE_RESULT, R"(file_biquad8_avx2.txt)");

	// ---------------------------------------------------------------------------------------------
#if (defined(__AVX512F__) && defined(__AVX512CD__) && defined(__AVX512DQ__) && defined(__AVX512BW__)) || defined(_MSC_VER)
	if (InstructionSet::AVX512F())
	{
		// zero history
		printf("\nbiquad8_avx512:\n");
		if (out) memset(out, 0, dataN * stride * sizeof(double));
		if (hist) memset(hist, 0, 2 * stride * sections * sizeof(double));

		// time biquad8_avx512
		before = clock::now();
		for (int k = 0; k < repeattest; ++k) 
		{
			for (int i = 0; i < dataN; ++i)
			{
				for (int d = 0; d < samples; ++d)
				{
					for (int j = 0; j <= stride - 8; j += 8)
					{
						biquad8_avx512(inp + (i * samples + d) * stride + j, out + i * stride + j, coeff8, hist + 2 * j * sections, sections);
					}
				}
			}
		}
		sec duration_biquadAVX512 = clock::now() - before;
		printf("It took %g sec\n", duration_biquadAVX512.count());
		printf("It was %gx faster\n", duration0.count() / duration_biquadAVX512.count());

#ifdef CHECK_RESULT
		diff = check_result(out, outfirst, dataN * stride);
		printf("Biggest deviation %g\n", diff);
#endif
		// write results
		write_results(out, MAX_WRITE_RESULT, R"(file_biquad8_avx512.txt)");
	}
#endif

	// ---------------------------------------------------------------------------------------------
	// zero history
	printf("\nbiquad_stride1_std:\n");
	if (out) memset(out, 0, dataN * stride * sizeof(double));
	if (hist) memset(hist, 0, 2 * stride * sections * sizeof(double));

	// time biquad_stride1_std
	before = clock::now();
	for (int k = 0; k < repeattest; ++k) {
			for (int i = 0; i < dataN; ++i)
			{
				biquad_stride1_std(inp + i * stride * samples, out + i * stride, coeff, hist, sections, stride, samples, 0);
			}
	}
	sec duration_biquad_stride1 = clock::now() - before;
	printf("It took %g sec\n", duration_biquad_stride1.count());
	printf("It was %gx faster\n", duration0.count() / duration_biquad_stride1.count());

#ifdef CHECK_RESULT
	diff = check_result(out, outfirst, dataN * stride);
	printf("Biggest deviation %g\n", diff);
#endif
	// write results
	write_results (out, MAX_WRITE_RESULT, R"(file_biquad_stride1_std.txt)");

	// ---------------------------------------------------------------------------------------------
	// zero history
	printf("\nbiquad_stride2_std:\n");
	if (out) memset(out, 0, dataN * stride * sizeof(double));
	if (hist) memset(hist, 0, 2 * stride * sections * sizeof(double));

	// time biquad_stride2_std
	before = clock::now();
	for (int k = 0; k < repeattest; ++k) {
		for (int i = 0; i < dataN; ++i)
		{
			biquad_stride2_std(inp + i * stride * samples, out + i * stride, coeff, hist, sections, stride, samples, 0);
		}
	}
	sec duration_iquad_stride2 = clock::now() - before;
	printf("It took %g sec\n", duration_iquad_stride2.count());
	printf("It was %gx faster\n", duration0.count() / duration_iquad_stride2.count());
#ifdef CHECK_RESULT
	diff = check_result(out, outfirst, dataN * stride);
	printf("Biggest deviation %g\n", diff);
#endif
	// write results
	write_results(out, MAX_WRITE_RESULT, R"(file_biquad_stride2_std.txt)");

	// ---------------------------------------------------------------------------------------------
	// zero history
	printf("\nbiquad_stride4_std:\n");
	if (out) memset(out, 0, dataN * stride * sizeof(double));
	if (hist) memset(hist, 0, 2 * stride * sections * sizeof(double));

	// time biquad_stride4_std
	before = clock::now();
	for (int k = 0; k < repeattest; ++k) {
		for (int i = 0; i < dataN; ++i)
		{
			biquad_stride4_std (inp + i * stride * samples, out + i * stride, coeff, hist, sections, stride, samples, 0);
		}
	}
	sec duration_biquad_stride4 = clock::now() - before;
	printf("It took %g sec\n", duration_biquad_stride4.count());
	printf("It was %gx faster\n", duration0.count() / duration_biquad_stride4.count());
#ifdef CHECK_RESULT
	diff = check_result(out, outfirst, dataN * stride);
	printf("Biggest deviation %g\n", diff);
#endif
	// write results
	write_results(out, MAX_WRITE_RESULT, R"(file_biquad_stride4_std.txt)");

	// ---------------------------------------------------------------------------------------------
	// zero history
	if (!InstructionSet::SSE2() || !InstructionSet::FMA()) return 1;
	printf("\nbiquad_stride2_section3_sse3:\n");
	if (out) memset(out, 0, dataN * stride * sizeof(double));
	if (hist) memset(hist, 0, 2 * stride * sections * sizeof(double));

	// time SSE2/FMA biquad
	before = clock::now();
	for (int k = 0; k < repeattest; ++k) {
		for (int i = 0; i < dataN; ++i)
		{
			biquad_stride2_section3_sse3(inp + i * stride * samples, out + i * stride, coeff, hist, stride, samples, 0);
		}
	}
	sec duration_biquad_sse3_stride2_section3 = clock::now() - before;
	printf("It took %g sec\n", duration_biquad_sse3_stride2_section3.count());
	printf("It was %gx faster\n", duration0.count() / duration_biquad_sse3_stride2_section3.count());
#ifdef CHECK_RESULT
	diff = check_result(out, outfirst, dataN * stride);
	printf("Biggest deviation %g\n", diff);
#endif

	// write results
	write_results(out, MAX_WRITE_RESULT, R"(file_biquad_stride2_section3_sse3.txt)");

	// ---------------------------------------------------------------------------------------------
	// zero history
	if (!InstructionSet::SSE2() || !InstructionSet::FMA()) return 1;
	printf("\nbiquad_stride8_section3_sse3:\n");
	if (out) memset(out, 0, dataN * stride * sizeof(double));
	if (hist) memset(hist, 0, 2 * stride * sections * sizeof(double));

	// time SSE2/FMA biquad
	before = clock::now();
	for (int k = 0; k < repeattest; ++k) {
		for (int i = 0; i < dataN; ++i)
		{
			biquad_stride8_section3_sse3(inp + i * stride * samples, out + i * stride, coeff, hist, stride, samples, 0);
		}
	}
	sec duration_biquad_sse3_stride8_section3 = clock::now() - before;
	printf("It took %g sec\n", duration_biquad_sse3_stride8_section3.count());
	printf("It was %gx faster\n", duration0.count() / duration_biquad_sse3_stride8_section3.count());
#ifdef CHECK_RESULT
	diff = check_result(out, outfirst, dataN * stride);
	printf("Biggest deviation %g\n", diff);
#endif

	// write results
	write_results(out, MAX_WRITE_RESULT, R"(file_biquad_stride8_section3_sse3.txt)");

	// ---------------------------------------------------------------------------------------------
	// zero history
	if (!InstructionSet::AVX2()) return 1;
	printf("\nbiquad_stride4_section3_avx2:\n");
	if (out) memset(out, 0, dataN * stride * sizeof(double));
	if (hist) memset(hist, 0, 2 * stride * sections * sizeof(double));

	// time AVX2 biquad
	before = clock::now();
	for (int k = 0; k < repeattest; ++k) {
		for (int i = 0; i < dataN; ++i)
		{
			biquad_stride4_section3_avx2 (inp + i * stride * samples, out + i * stride, coeff, hist, stride, samples, 0);
		}
	}
	sec duration_biquad_avx2_stride4_section3 = clock::now() - before;
	printf ("It took %g sec\n", duration_biquad_avx2_stride4_section3.count());
	printf ("It was %gx faster\n", duration0.count() / duration_biquad_avx2_stride4_section3.count());
#ifdef CHECK_RESULT
	diff = check_result(out, outfirst, dataN * stride);
	printf("Biggest deviation %g\n", diff);
#endif

	// write results
	write_results(out, MAX_WRITE_RESULT, R"(file_biquad_stride4_section3_avx2.txt)");

	// ---------------------------------------------------------------------------------------------
	// zero history
	printf("\nbiquad_stride8_section3_avx2:\n");
	if (out) memset(out, 0, dataN * stride * sizeof(double));
	if (hist) memset(hist, 0, 2 * stride * sections * sizeof(double));

	// time AVX2 biquad
	before = clock::now();
	for (int k = 0; k < repeattest; ++k) {
		for (int i = 0; i < dataN; ++i)
		{
			biquad_stride8_section3_avx2 (inp + i * stride * samples, out + i * stride, coeff, hist, stride, samples, 0);
		}
	}
	sec duration_biquad_avx2_stride8_section3 = clock::now() - before;
	printf("It took %g sec\n", duration_biquad_avx2_stride8_section3.count());
	printf("It was %gx faster\n", duration0.count() / duration_biquad_avx2_stride8_section3.count());
#ifdef CHECK_RESULT
	diff = check_result(out, outfirst, dataN * stride);
	printf("Biggest deviation %g\n", diff);
#endif

	// write results
	write_results(out, MAX_WRITE_RESULT, R"(file_biquad_stride8_section3_avx2.txt)");

	// ---------------------------------------------------------------------------------------------
	// zero history
	printf("\nbiquad_stride16_section3_avx2:\n");
	if (out) memset(out, 0, dataN * stride * sizeof(double));
	if (hist) memset(hist, 0, 2 * stride * sections * sizeof(double));

	// time AVX2 biquad
	before = clock::now();
	for (int k = 0; k < repeattest; ++k) {
		for (int i = 0; i < dataN; ++i)
		{
			biquad_stride16_section3_avx2 (inp + i * stride * samples, out + i * stride, coeff, hist, stride, samples, 0);
		}
	}
	sec duration_biquad_avx2_stride16_section3 = clock::now() - before;
	printf("It took %g sec\n", duration_biquad_avx2_stride16_section3.count());
	printf("It was %gx faster\n", duration0.count() / duration_biquad_avx2_stride16_section3.count());
#ifdef CHECK_RESULT
	diff = check_result(out, outfirst, dataN * stride);
	printf("Biggest deviation %g\n", diff);
#endif

	// write results
	write_results(out, MAX_WRITE_RESULT, R"(file_biquad_stride16_section3_avx2.txt)");

	// ---------------------------------------------------------------------------------------------
	// zero history
	printf("\nbiquad_stride32_section3_avx2:\n");
	if (out) memset(out, 0, dataN * stride * sizeof(double));
	if (hist) memset(hist, 0, 2 * stride * sections * sizeof(double));

	// time AVX2 biquad
	before = clock::now();
	for (int k = 0; k < repeattest; ++k) {
		for (int i = 0; i < dataN; ++i)
		{
			biquad_stride32_section3_avx2(inp + i * stride * samples, out + i * stride, coeff, hist, stride, samples, 0);
		}
	}
	sec duration_biquad_avx2_stride32_section3 = clock::now() - before;
	printf("It took %g sec\n", duration_biquad_avx2_stride32_section3.count());
	printf("It was %gx faster\n", duration0.count() / duration_biquad_avx2_stride32_section3.count());
#ifdef CHECK_RESULT
	diff = check_result(out, outfirst, dataN * stride);
	printf("Biggest deviation %g\n", diff);
#endif

	// write results
	write_results(out, MAX_WRITE_RESULT, R"(file_biquad_stride32_section3_avx2.txt)");

#if (defined(__AVX512F__) && defined(__AVX512CD__) && defined(__AVX512DQ__) && defined(__AVX512BW__)) || defined(_MSC_VER)
	if (!InstructionSet::AVX512F()) return 1;
	// ---------------------------------------------------------------------------------------------
	// zero history
	printf("\nbiquad_stride8_section3_avx512:\n");
	if (out) memset(out, 0, dataN * stride * sizeof(double));
	if (hist) memset(hist, 0, 2 * stride * sections * sizeof(double));

	// time AVX2 biquad
	before = clock::now();
	for (int k = 0; k < repeattest; ++k) {
		for (int i = 0; i < dataN; ++i)
		{
			biquad_stride8_section3_avx512(inp + i * stride * samples, out + i * stride, coeff, hist, stride, samples, 0);
		}
	}
	sec duration_biquad_avx512_stride8_section3 = clock::now() - before;
	printf("It took %g sec\n", duration_biquad_avx512_stride8_section3.count());
	printf("It was %gx faster\n", duration0.count() / duration_biquad_avx512_stride8_section3.count());
#ifdef CHECK_RESULT
	diff = check_result(out, outfirst, dataN * stride);
	printf("Biggest deviation %g\n", diff);
#endif

	// write results
	write_results(out, MAX_WRITE_RESULT, R"(file_biquad_stride8_section3_avx512.txt)");

	// ---------------------------------------------------------------------------------------------
	// zero history
	printf("\nbiquad_stride16_section3_avx512:\n");
	if (out) memset(out, 0, dataN * stride * sizeof(double));
	if (hist) memset(hist, 0, 2 * stride * sections * sizeof(double));

	// time AVX2 biquad
	before = clock::now();
	for (int k = 0; k < repeattest; ++k) {
		for (int i = 0; i < dataN; ++i)
		{
			biquad_stride16_section3_avx512(inp + i * stride * samples, out + i * stride, coeff, hist, stride, samples, 0);
		}
	}
	sec duration_biquad_avx512_stride16_section3 = clock::now() - before;
	printf("It took %g sec\n", duration_biquad_avx512_stride16_section3.count());
	printf("It was %gx faster\n", duration0.count() / duration_biquad_avx512_stride16_section3.count());
#ifdef CHECK_RESULT
	diff = check_result(out, outfirst, dataN * stride);
	printf("Biggest deviation %g\n", diff);
#endif

	// write results
	write_results(out, MAX_WRITE_RESULT, R"(file_duration_biquad_avx512_stride16_section3.txt)");

	// ---------------------------------------------------------------------------------------------
	// zero history
	printf("\nbiquad_stride32_section3_avx512:\n");
	if (out) memset(out, 0, dataN * stride * sizeof(double));
	if (hist) memset(hist, 0, 2 * stride * sections * sizeof(double));

	// time AVX2 biquad
	before = clock::now();
	for (int k = 0; k < repeattest; ++k) {
		for (int i = 0; i < dataN; ++i)
		{
			biquad_stride32_section3_avx512(inp + i * stride * samples, out + i * stride, coeff, hist, stride, samples, 0);
		}
	}
	sec duration_biquad_avx512_stride32_section3 = clock::now() - before;
	printf("It took %g sec\n", duration_biquad_avx512_stride32_section3.count());
	printf("It was %gx faster\n", duration0.count() / duration_biquad_avx512_stride32_section3.count());
#ifdef CHECK_RESULT
	diff = check_result(out, outfirst, dataN * stride);
	printf("Biggest deviation %g\n", diff);
#endif

	// write results
	write_results(out, MAX_WRITE_RESULT, R"(file_biquad_stride32_section3_avx512.txt)");
#endif
}
